/*	License
 * 		inside EU: EUPL
 * 		outside EU: GPL or EUPL
 *  The program comes AS IS and without any warranty unless required by law
 *  (C) 2012 thecodemonkey
 */

package gui.settings.dialog.group;

import gui.resource.ResourceManager;
import gui.settings.dialog.component.SettingsComponent;

import java.util.LinkedList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JPanel;

import settings.Settings;
import settings.SettingsLoadable;


public abstract class SettingsComponentGroup extends JPanel implements SettingsLoadable {
	private static final long serialVersionUID = 1L;
	
	protected SettingsComponent singleComponent;
	protected List<SettingsComponent> components;
	private String title;
	
	public SettingsComponentGroup(String title,ResourceManager rm) {
		this(title, rm, true);
	}
	
	public SettingsComponentGroup(String title,ResourceManager rm,boolean autoCreate) {
		this.title = title;
		setBorder(BorderFactory.createTitledBorder(title));
		if (autoCreate) createSettingsComponents(rm);
	}
	
	protected abstract void createSettingsComponents(ResourceManager rm);
	
	
	protected void addSettingsComponent(SettingsComponent comp) {
		addSettingsComponent(comp, null);
	}
	
	protected void addSettingsComponent(SettingsComponent comp,JComponent parent, Object layoutConstraints) {
		addToCompList(comp);
		parent.add(comp.getComponent(),layoutConstraints);
	}
	
	protected void addSettingsComponent(SettingsComponent comp, Object layoutConstraints) {
		addToCompList(comp);
		add(comp.getComponent(),layoutConstraints);
	}
	
	private void addToCompList(SettingsComponent comp) {
		if (singleComponent==null&&components==null) {
			singleComponent = comp;
		} else {
			if (components==null) {
				components = new LinkedList<>();
				components.add(singleComponent);
				singleComponent=null;
			} 
			components.add(comp);
		}
	}
	
	/**
	 * saves the settings for the group 
	 */
	@Override
	public void saveSettings(Settings s ) {
		if (singleComponent!=null) {
			singleComponent.saveSettings(s);
		} else {
			for (SettingsComponent c:components) {
				c.saveSettings(s);
			}
		}
	}
	
	@Override
	public void loadSettings(Settings s) {
		try {
		if (singleComponent!=null) {
			singleComponent.loadSettings(s);
		} else if (components!=null){
			for (SettingsComponent c:components) {
				c.loadSettings(s);
			}
		} else {
			System.out.println("Fehler: components=null");
		}
		} catch (Exception e) {
			System.out.println("Error load settings "+title);
		}
	}
	
	@Override
	public String toString() {
		return title;
	}
}
