package model.transfer;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import shoshone.httpclient.ShoshoneClient;
import util.exception.CancelException;
import core.io.Application;

public class ConfigUpdate {
	
//	public static boolean downloadUpdates(String name,String type,Application app,Path cfgDir,boolean force) throws IOException, CancelException{
//		return downloadUpdates(name, type, app, cfgDir, null, force);
//	}
	
	public static boolean downloadUpdates(String name,String type,Application app,Path cfgDir,boolean force) throws IOException, CancelException{
		String t="revision"+capitalize(type);
		String url="http://thecodemonkey.bplaced.net/get_cfgs.php?name="+name+'.'+type;
		String version= app.getVersion();
		if (version!=null) url+="&version="+version;
		ShoshoneClient sc=new ShoshoneClient(url);
		if (!force) sc.addRequestProperty("If-Modified-Since",getCurrentRevisionDate(app,t));
		sc.applyGetRequest(false);
		
		if (sc.getResponseCode()==200){
			sc.readResponseAndSave(cfgDir.resolve(sc.getFilename()));
			app.updateManifest(t, sc.getResponseHeader("Last-Modified"));
			return true;
		} else if (sc.getResponseCode()==304){
			return false;
		} else {
			throw new IOException();
		}
		
	}
	
	public static boolean downloadNonXMLUpdates(String name,String type,Application app,Path outDir,boolean force) throws ZipException, IOException, CancelException {
		boolean result=downloadUpdates(name, type, app, outDir,  force);
		if (result) {
			Path zf = outDir.resolve("additional.zip");
			ZipFile zip=new ZipFile(zf.toFile());
			Enumeration<? extends ZipEntry> en=zip.entries();
			while (en.hasMoreElements()) {
				ZipEntry e = en.nextElement();
				if (!e.isDirectory()) {
					OutputStream out=new BufferedOutputStream(Files.newOutputStream(outDir.resolve(e.getName()) ));
					InputStream entryIn = new BufferedInputStream(zip.getInputStream(e));
					int b;
					while ((b=entryIn.read())>-1){
						out.write(b);
					}
					out.close();
				}
			}
			zip.close();
			Files.delete(zf);
		}
		return result;
	}
	
	private static String capitalize(String s ) {
		return Character.toUpperCase(s.charAt(0)) +s.substring(1);
	}
	
		
	private static String getCurrentRevisionDate(Application app,String t){
		String d= app.getManifestValue(t);
		if (d==null||d.isEmpty())
			return "Sun, 01 Jan 2012 00:00:00 GMT";
		else
			return d;
	}
}
