/*	License
 * 		inside EU: EUPL
 * 		outside EU: GPL or EUPL
 *  The program comes AS IS and without any warranty unless required by law
 *  (C) 2012 thecodemonkey
 */

package shoshone.httpclient;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;

import util.exception.CancelException;

public class MultipartFile extends Multipart {
	
	private Path file;
	private String mime;
	
	public MultipartFile(Shoshone urlConn, String fieldName,String mime,Path file) {
		super(urlConn, fieldName);
		this.file=file;
		this.mime=mime;
		
	}

	@Override
	public void apply() throws IOException, CancelException {
		BufferedInputStream fis=null;
		try {
			urlConn.writeString(PREFIX);
			urlConn.writeStringAndNewLine(boundary); //2
			
			urlConn.writeStringAndNewLine(CONTENT_DISP+fieldName+"\"; filename=\""+file.getFileName()+"\""); //2
			urlConn.writeStringAndNewLine("Content-Type: "+mime); //2
			
			urlConn.writeNewline(); //2
			
			fis=new BufferedInputStream(Files.newInputStream(file, StandardOpenOption.READ));
			
			int r;
			while ( (r=fis.read())>-1) {
				urlConn.writeByte(r);
			}
			
			urlConn.writeNewline(); //2
		} catch (IOException e) {
			e.printStackTrace();
			throw e;
		} finally {
			fis.close();
		}
	}

	@Override
	public long getContentLength() throws IOException {
		return PREFIX.length()+boundary.length()+
		(CONTENT_DISP+fieldName+"\"; filename=\""+file.getFileName()+"\"").length()
		+("Content-Type: "+mime).length()
		+Files.size(file)
		+10
		;
	}
}
