/*	License
 * 		inside EU: EUPL
 * 		outside EU: GPL or EUPL
 *  The program comes AS IS and without any warranty unless required by law
 *  (C) 2012 thecodemonkey
 *  
 *  Version 1.9
 */

package shoshone.httpclient;
import static shoshone.httpclient.Multipart.PREFIX;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketException;
import java.net.URL;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.OpenOption;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.zip.GZIPInputStream;

import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;

import net.util.Proxy;

import util.StringUtil;
import util.exception.CancelException;

public class ShoshoneClient extends Shoshone {
	
	private final static String CRLF = "\r\n";

	private Socket socket;
	private final static Charset UTF8=Charset.forName("UTF-8");
	
	private BufferedInputStream in = null;
	private BufferedOutputStream out=null;
	
	private final static String GET_METHOD="GET ",POST_METHOD="POST ";
	
	private String boundary;
	
	private URL url;

	private boolean expect100continue,enableCookies ;
	
	private LinkedHashMap<String, String> requestProperties;
	private HashMap<String, String> responseHeaders;
	private LinkedList<Multipart> multipart;
	
	private LinkedHashSet<Cookie> cookies;
	
	private String response;
	
	private int responseCode=-1;

	private long bytes;
	
	private boolean followRedirects,resetDone,autoReferer,gzip;
	
	private InetAddress inetAddress;
	
	private final static int MAX_REDIRECTS=2,BUFFER=8192;//16384;
	
	private int redirects;
	
	private final static String BREAK_READ="</html>";
	
	private long contentLength;
	private String headerFilename,type=null;
	
	private boolean chunked;
	private boolean cancel;
	private boolean gzipEnabled = true;
	private static char[] ILLEGAL_FILENAME_CHARS={'?','<','>','\'','\\','/','*','|',':','"'}; 
	
	private Proxy proxy;
	private boolean active;
	
	public ShoshoneClient() throws IOException{
		this(null,true,Proxy.noProxy());
	}
	
	public ShoshoneClient(String urlString) throws IOException {
		this(urlString,true,Proxy.noProxy());
	}
	
	public ShoshoneClient(String urlString,Proxy proxy) throws IOException {
		this(urlString,true,proxy);
	}
	
	public ShoshoneClient(String urlString,boolean enableCookies) throws IOException {
		this(urlString,enableCookies,Proxy.noProxy());
	}
	
	public ShoshoneClient(String urlString,boolean enableCookies, Proxy proxy) throws IOException {
		this.enableCookies=enableCookies;
		this.proxy=proxy;
		
		requestProperties=new LinkedHashMap<String, String>();
		
		multipart=new LinkedList<Multipart>();
		cookies=new LinkedHashSet<Cookie>();
		responseHeaders=new HashMap<>();
		redirects=MAX_REDIRECTS;
		reset(urlString) ;
	}
	
	private void reset(String urlString) throws IOException {
		reset(urlString, true,true);
	}
	
	private void reset(String urlString,boolean resetRedirects,boolean resetRequestProperties) throws IOException {
		if (urlString!=null) {
			resetDone=true;
			bytes=0;
			if (in!=null)
				closeConnection();
			
			URL newUrl=new URL(urlString);
			String referer=null;
			if (url==null) {
				inetAddress=InetAddress.getByName(newUrl.getHost());
			} else if (!newUrl.getHost().equals(url.getHost())) {
				inetAddress=InetAddress.getByName(newUrl.getHost());
				referer=url.toExternalForm();
			} else {
				referer=url.toExternalForm();
			}
			
			url=newUrl;
			
			contentLength=-1L;
			chunked=false;
			gzip=false;
	        createSocket();
	        in = new BufferedInputStream(socket.getInputStream(),BUFFER);
			out=new BufferedOutputStream(socket.getOutputStream(),BUFFER);
	
			
			expect100continue=false;
			
			boundary=null;
			
			multipart.clear();
			
			if (resetRedirects) {
				redirects=MAX_REDIRECTS; 
				
			}
			
			if (resetRequestProperties) {
				setDefaultRequestProperties();
			}
			requestProperties.put("Host", url.getHost());
			if (autoReferer && referer!=null) requestProperties.put("Referer", referer);
			responseHeaders.clear();
		}
	}
	
	/**
	 * prepare multipart/form data request
	 * 
	 * the method caller must set the boundary first
	 * 
	 */
	public void addMultipartFormData(Multipart m) {
		
		if (boundary==null) throw new IllegalStateException("Boundary is not set");
		
		m.setBoundary(boundary);
		multipart.add(m);
	}
	
	
	public synchronized void applyGetRequest() throws IOException, CancelException {
		applyGetRequest(true);
	}
			
	public synchronized void applyGetRequest(boolean readResponseBody) throws IOException, CancelException {
		try{
			cancel=false;
			setActive();
			if (!resetDone) reset(url.toExternalForm(),true,false);
	//		createProxyTunnel();
			resetDone=false;
			writeStringAndNewLine(getMethodString(GET_METHOD));
			
			for (String requestProperty:requestProperties.keySet()) {
				writeStringAndNewLine(requestProperty+": "+requestProperties.get(requestProperty));
			}
			
			writeNewline();
			writeNewline();
			out.flush();
			
			readResponseHeader(readResponseBody);
		} catch (IOException e){
			if (cancel) {
				throwException();
			}else{
				throw e;
			}
		} finally{
			setInactive();
		}
//		unlock();
	}
	
//	private void unlock(){
//		synchronized (cancelLock) {
//			cancelLock.notify();
//		};
//	}
	
	
	public synchronized void applyPostRequest(List<PostParameter> params) throws IOException, CancelException {
		applyPostRequest(params, true);
	}
	
	public synchronized void applyPostRequest(List<PostParameter> params,boolean readResponseBody) throws IOException, CancelException {
		try{
			setActive();
			if (!resetDone) reset(url.toExternalForm(),true,false);
	//		createProxyTunnel();
			
			writeStringAndNewLine(getMethodString(POST_METHOD));
			
			requestProperties.put("Content-Type" ,"application/x-www-form-urlencoded");
			
			for (String requestProperty:requestProperties.keySet()) {
				writeStringAndNewLine(requestProperty+": "+requestProperties.get(requestProperty));
			}
			
			StringBuilder sb=new StringBuilder();
			
			Iterator<PostParameter> it=params.iterator();
			
			while (it.hasNext()) {
				sb.append(it.next());
				if (it.hasNext()) {
					sb.append('&');
				}
			}
			
			String parameters = sb.toString();
			
			writeStringAndNewLine("Content-Length: "+parameters.length());
			writeNewline();
			
			writeString(parameters);
			
			out.flush();
			resetDone=false;
			requestProperties.remove("Content-Type");
			readResponseHeader(readResponseBody);
		} catch (IOException e) {
			if (cancel) {
				throwException();
			}else{
				throw e;
			}
		} finally{
			setInactive();
		}
//		unlock();
	}
	
	private void readResponseHeader(boolean readResponseBody) throws IOException, CancelException {
		RequestDelimiter d=new RequestDelimiter();
		
		StringBuilder sb=new StringBuilder();
		int r;
		while ( (r=in.read())>-1) {
			sb.append((char)r);
			
			if (d.isDelimiterReached(r)) {
				break;
			} 
		}
		
		String[] hdr=sb.toString().split(CRLF);
		
		String followRedirectsUrl=null;
		for (String line:hdr) {
			String lineLower=line.toLowerCase();
			if (enableCookies &&lineLower.startsWith("set-cookie")) {
				Cookie c=new Cookie(line.substring(12));
				if (cookies.contains(c)) {
					cookies.remove(c);
				}
				addCookie(c);
			} else if (lineLower.startsWith("http/1.")) {
				responseCode=Integer.parseInt(line.substring(9, 12));
			} else if (lineLower.startsWith("content-length")) {
				contentLength=Long.parseLong(line.substring(16));
			} else if (lineLower.startsWith("content-disposition")) {
				int k=line.indexOf("filename=\"");
				if (k>-1) {
					headerFilename=line.substring(k+10,line.lastIndexOf('\"'));
				} else {
					k=line.indexOf("filename=");
					
					if (k>-1) {
						int j=line.lastIndexOf(';');
						if (j>k)
							headerFilename=line.substring(k+9,j);
						else
							headerFilename=line.substring(k+9);
					}
				}
			} else if (lineLower.startsWith("content-type")) {
				int k=line.lastIndexOf('/');
				if (k>-1) {
					int sc = line.indexOf(';',k+1);
					if (sc==-1)
						type=line.substring(k+1);
					else
						type=line.substring(k+1, sc);
				} else {
					type=null;
				}
			} else if (followRedirects && redirects>0 && lineLower.startsWith("location")) {
				
				String location=line.substring(10);
				followRedirectsUrl=UrlUtil.getAbsolutePath(location, getUrl());
				
			} else if (lineLower.startsWith("transfer-encoding")) {
				chunked=line.substring(19, 26).equals("chunked");
			} else if (lineLower.startsWith("content-encoding")) {
				gzip = line.substring(18, 22).equals("gzip");
			} else if (lineLower.startsWith("refresh")) {
				followRedirectsUrl=UrlUtil.getAbsolutePath(line.substring(line.indexOf("url=")+4),getUrl());
				
			} else {
				int k=line.indexOf(':');
				if (k==-1) {
					System.out.println(line);
				}
				String hdrName=line.substring(0,k);
				String hdrVal=line.substring(k+2);
				responseHeaders.put(hdrName, hdrVal);
			}
		}
		
		if (followRedirectsUrl!=null) {
			followRedirect(followRedirectsUrl,readResponseBody);
		} else if (readResponseBody) {
			readResponseBody();
		}
	}
	
	private void throwException() throws CancelException{
		bytes=0;
		synchronized (cancelLock) {
			cancelLock.notify();
		}
		throw new CancelException();
	}
	
	private void writeFile(BufferedOutputStream fos) throws IOException{
		int r;
		if (contentLength==-1L) {
			while ((r=in.read())>-1) {
				fos.write(r);
				bytes++;
			}
		} else {
			int c=0;
			while (c<contentLength&&(r=in.read())>-1) {
				fos.write(r);
				bytes++;
				c++;
			}
//			System.out.println(c);
		}
	}
	
	private void decompressGzip(Path tmpFile,BufferedOutputStream fos) throws IOException {
		BufferedInputStream in=	new BufferedInputStream(new GZIPInputStream(Files.newInputStream(tmpFile),BUFFER));
		int r;
		while ((r=in.read())>-1) {
			fos.write(r);
		}
		in.close();
	}
	
	private void writeFileChunk(int length, BufferedOutputStream fos) throws IOException{
		int r;
		int c=0;
		while (c<length&&(r=in.read())>-1) {
			fos.write(r);
			bytes++;
			c++;
		}
	}
	
	
	synchronized public void readResponseAndSave(Path f ) throws IOException, CancelException {
		setActive();
		OpenOption[] options= new OpenOption[]{StandardOpenOption.CREATE,StandardOpenOption.WRITE,StandardOpenOption.TRUNCATE_EXISTING};
		BufferedOutputStream fos=new BufferedOutputStream(Files.newOutputStream(f, options));;
		
			if (gzip) {
				Path tmpFile=null;
				try {
					tmpFile=Files.createTempFile(Long.toString(System.currentTimeMillis()),f.getFileName().toString());
					BufferedOutputStream fosTemp=new BufferedOutputStream(Files.newOutputStream(tmpFile, options));
					if (chunked) {
						int b;
						
						while ((b=in.read())>-1) {
							int chunk=readChunkSize(b);
							if (chunk==0) break;
							writeFileChunk(chunk, fosTemp);
							in.skip(2);
						}
						fosTemp.close();
					} else {
						writeFile(fosTemp);
						fosTemp.close();
					}
					
					decompressGzip(tmpFile,fos);
				} catch (IOException e) {
					if (cancel) {
						throwException();
					}else{
						throw e;
					}
				} finally {
					
					fos.close();
					if (contentLength>-1L && Files.size(tmpFile)<contentLength){
						Files.deleteIfExists(tmpFile);
						setInactive();
						throw new IOException();
					}
					Files.deleteIfExists(tmpFile);
					setInactive();
				}
				
			} else {
				try {
					if (chunked) {
						int b;
						while ((b=in.read())>-1) {
							int chunk=readChunkSize(b);
							if (chunk==0) break;
							writeFileChunk(chunk, fos);
							in.skip(2);
						}
					} else {
						writeFile(fos);
					}
				} catch (IOException e) {
					if (cancel) {
						throwException();
					}else{
						throw e;
					}
				} finally {
					fos.close();
					if (contentLength>-1L && Files.size(f)<contentLength){
						setInactive();
						throw new IOException();
					}
					setInactive();
				}
			}
			
		
	}
	
	private void readResponseBody() throws IOException, CancelException {
		StringBuilder sbResponse=new StringBuilder();
		try {
			if (gzip) {
				if (chunked) {
					int b;
					DynamicUByteBuffer buf=new DynamicUByteBuffer();
					while ((b=in.read())>-1) {
						int chunk=readChunkSize(b);
						if (chunk==0) break;
						buf.put(readGzip(chunk));
						
						in.skip(2);
					}
					gzipToResponse(sbResponse, buf.getArr());
					
				} else {
					gzipToResponse(sbResponse, readGzip((int) contentLength));
				}
			} else {
				if (chunked) {
					int b;
					while ((b=in.read())>-1) {
						int chunk=readChunkSize(b);
						if (chunk==0) break;
						readPlaintext(sbResponse, chunk);
						in.skip(2);
					}
				} else {
					readPlaintext(sbResponse, (int)contentLength);
				}
			}
			response=sbResponse.toString();
		} catch (IOException e) {
			if (cancel) {
				throwException();
			}else{
				throw e;
			}
		}
	}

	private int addChunkChar(int currentVal,char c) {
		if (c >= '1' && c<='9')
			return (currentVal<<4)+((int)c)-48;
		else if (c >= 'a' && c <= 'f')
			return (currentVal<<4)+((int)c)-87;
		else if (c=='0'){
			return currentVal<<4;
		}
		return currentVal;
	}
	
	public void applyMultipartRequest() throws IOException, CancelException {
		cancel=false;
		setActive();
		try{
		if (!resetDone) reset(url.toExternalForm());
//		createProxyTunnel();
		
		long cl=0;
		for (Multipart mp:multipart) {
			cl+=mp.getContentLength();
		}
		cl+=getFinalizeMultipartLength();
		
		writeStringAndNewLine(getMethodString(POST_METHOD));
		writeStringAndNewLine("Content-Length: "+cl);
		
		addRequestProperty("Content-Type", "multipart/form-data; boundary="+boundary);
		
		if (expect100continue) {
			writeStringAndNewLine("Expect: 100-continue");
		}
		
		for (String requestProperty:requestProperties.keySet()) {
			writeStringAndNewLine(requestProperty+": "+requestProperties.get(requestProperty));
		}
				
		writeNewline();
		
		if (expect100continue) {
			
			out.flush();
	
			int read = in.read();
			
			StringBuilder sb=new StringBuilder();
			
			while (read != -1) {
				sb.append((char)read);
				
				String s=sb.toString();
				
				if (s.equals("HTTP/1.1 100 Continue\r\n\r\n")) {
					break;
				} else if (s.startsWith("HTTP/1.1 4")) {
					in.read();
					in.read();
					in.read();
					responseCode=Integer.parseInt(s.substring(9, 12));
					break;
				}
				
				read = in.read();
			}
		}
		
		for (Multipart mp:multipart) {
			
			mp.apply();
			out.flush();
		}
		finalizeMultipart();
		out.flush();
		
		readResponseHeader(true);

		
//		unlock();
		
		} catch (IOException e) {
			if (cancel) {
				throwException();
			} else {
				throw e;
			}
		} finally {
			setInactive();
			resetDone=false;
		}
	}
	
	private void finalizeMultipart() throws IOException {
		writeString(PREFIX);
		writeStringAndNewLine(boundary+PREFIX);
	}
	
	private int getFinalizeMultipartLength() {
		return boundary.length()+6;
	}
	
	public String getResponse() throws IOException {
		return response;
	}
	
	public void createBoundary() {
		boundary= "----------"+Long.toHexString(System.currentTimeMillis());
	}
	
	public void closeConnection() throws IOException {
//		in.close();
//		out.close();
		socket.close();
	}
	
	private String getMethodString(String method) {
		String path=null;
		if (proxy.getPort()==-1) {
			path=url.getFile();
			if (path.length()==0) {
				path="/";
			}
		} else {
			path=url.toExternalForm();
		}
		return method+path+" HTTP/1.1" ;
	}
	
	private void setDefaultRequestProperties() {
		requestProperties.clear();
		requestProperties.put("User-Agent", "Mozilla/5.0 (Windows NT 5.1; rv:12.0) Gecko/20100101 Firefox/12.0");
		requestProperties.put("Connection", "Keep-Alive, TE");
		requestProperties.put("TE", "deflate, gzip, chunked, identity, trailers");
		if (gzipEnabled) requestProperties.put("Accept-Encoding", "gzip");
		if (cookies!=null && !cookies.isEmpty() ) requestProperties.put("Cookie", getCookieString(cookies));
	}
	
	
	protected void writeString(String s) throws IOException {
		out.write(s.getBytes(UTF8));
	}
	
	protected void writeStringAndNewLine(String s) throws IOException {
		out.write(s.getBytes(UTF8));
		writeNewline();
	}
	
	protected void writeNewline() throws IOException {
		out.write(13);
		out.write(10);
	}
	
	public void setUrl(String urlString) throws IOException {
		reset(urlString);
	}
	
	private void followRedirect(String url,boolean readRsp) throws IOException, CancelException {
		reset(url,false,false);
//		addCookies(cookies);
		redirects--;
		applyGetRequest(readRsp);
		
	}
	
	public void clearCookies() {
		cookies.clear();
	}
	
	public Set<Cookie> getCookies() throws IOException {
		return cookies;
	}

	public int getResponseCode() {
		return responseCode;
	}

	protected void writeByte(int b) throws IOException, CancelException {
		try{
			out.write(b);
			bytes++;
		} catch (IOException e) {
			e.printStackTrace();
			if (cancel){
				throwException();
			} else {
				throw e;
			}
		}
	}
	
	public void addRequestProperty(RequestProperty rp) {
		addRequestProperty(rp.getName(), rp.getValue());
	}
	
	public void addRequestProperty(String name,String value) {
		
		if (name.equals("Expect") && value.equals("100-continue")) {
			expect100continue=true;
		} else if (name.equals("Cookie")) {
			cookies.add(new Cookie(value));
			requestProperties.put("Cookie", getCookieString(this.cookies));
		} else {
			requestProperties.put(name, value);
		}
	}
	
	public void addCookie(Cookie c) {
		cookies.add(c);
		requestProperties.put("Cookie", getCookieString(this.cookies));
	}
	
	public void addCookies(Set<Cookie> cookies) {
		if (cookies!=null && !cookies.isEmpty()) {
			if (cookies.size()>0) {
				this.cookies.addAll(cookies);
				requestProperties.put("Cookie", getCookieString(this.cookies));
			}
		}
	}

	private String getCookieString(Set<Cookie> cookies) {
		if (cookies!=null) {
			StringBuilder sb=new StringBuilder();
			
			Iterator<Cookie> it=cookies.iterator();
			while (true) {
				sb.append(it.next());
				if (it.hasNext()) {
					sb.append("; ");
				} else {
					break;
				}
			}
			return sb.toString();
		} else {
			return null;
		}
	}

	public long getTransferedBytes() {
		return bytes;
	}
	
	public void setFollowRedirects(boolean b) {
		followRedirects=b;
	}

	public String getUrl() {
		return url.toExternalForm();
	}

	public String getFilename() {
		if (headerFilename!=null && !headerFilename.isEmpty()) {
			return headerFilename;
		} else if (!url.getFile().isEmpty()) {
			
			String fn=StringUtil.replaceAll(url.getFile().substring(url.getFile().lastIndexOf('/')+1),":", "-");
			if (containsIllegalChar(fn)) {
				
				if (type==null) {
					return "file";
				} else {
					return "file."+type;
				}
			} else {
				return fn;
			}
		} else {
			return "file";
		}
	}
	
	public long getContentLength() {
		return contentLength;
	}
	
	public void setAutoReferer() {
		this.autoReferer = true;
	}
	
	public void setTimeout(int sec) {
		try {
			socket.setSoTimeout(sec*1000);
		} catch (SocketException e) {
			e.printStackTrace();
		}
	}
	
//	private void createProxyTunnel() throws IOException {
//		if (proxy.getHost()!=null) {
//			PrintWriter out=new PrintWriter(this.sslSocket.getOutputStream());
//			out.write(("CONNECT "+url.getHost()+":443 HTTP/1.1" + CRLF + "Host: "+url.getHost()+":443" + CRLF));
//	        out.write(CRLF);
//	        out.close();
//		}
//	}

	private void createSocket() throws IOException {
		
		if (proxy.getPort()==-1) {
			if (url.getProtocol().equals("http")){
				socket= new Socket(inetAddress, 80);
			} else if (url.getProtocol().equals("https")){
				socket =(SSLSocket) SSLSocketFactory.getDefault().createSocket(inetAddress, 443);
			}
			
		} else {
			socket= new Socket(proxy.getHost(), proxy.getPort());

//			sslSocket=(SSLSocket) SSLSocketFactory.getDefault().createSocket(proxy.getHost(), proxy.getPort());
		}
	}
	
	public static boolean containsIllegalChar(String fn) {
		if (fn.isEmpty()) {
			return true; 
		} else {
			for (char c:ILLEGAL_FILENAME_CHARS ) {
				for (int i=0;i<fn.length();i++) {
					if (fn.charAt(i)==c) return true;
				}
			}
			return false;
		}
	}
	
	public synchronized void applyGetRequest(String url) throws IOException, CancelException {
		setUrl(url);
		applyGetRequest();
	}
	
	public void setProxy(Proxy proxy) {
		this.proxy=proxy;
	}
	
	private Object cancelLock=new Object();
	
	public void cancel() {
	
		synchronized (cancelLock) {
			cancel=true;
			
			try {
				if (active) {
					closeConnection();
					cancelLock.wait();
				}
			} catch (InterruptedException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		System.out.println("cancelled");
	}
	
	@Override
	protected void finalize() throws Throwable {
		try {
			closeConnection();
		} catch (IOException e) {}
	}
	
	public String getResponseHeader(String name) {
		return responseHeaders.get(name);
	}
	
	public Map<String, String> getResponseHeaders() {
		return Collections.unmodifiableMap( responseHeaders);
	}
	
	
	private int[] readGzip(int length) throws IOException {
		if (length>-1){ 
			int[] gzipped=new int[length];
			int i=0;
			int r;
			while (i<length&&(r=in.read())>-1 ) {
				gzipped[i++]=r;
				bytes++;
			}
			return gzipped;
		} else {
			DynamicUByteBuffer arr=new DynamicUByteBuffer();
			int r;
			while ((r=in.read())>-1 ) {
				arr.put((byte) r);
				bytes++;
			}
			return arr.getArr();
		}
	}
	
	private void gzipToResponse(StringBuilder sbResponse, int[] gzipped) throws IOException {
		int c=0;
		
		GZIPInputStream g=new GZIPInputStream(new UByteArrayInputStream(gzipped));
		int r;
		try{
		while ( (r=g.read())>-1) {
			sbResponse.append((char)r);
			c++;
		}
		} catch (Exception e){
			System.out.println(c);
			e.printStackTrace();
		}
	}	
	
	private void readPlaintext(StringBuilder sbResponse, int length) throws IOException {
		int r;
		
		if (length==-1) {
			while ((r=in.read())>-1){
				sbResponse.append((char)r);
				bytes++;
				if (sbResponse.lastIndexOf(BREAK_READ)>-1) break;
			}
		} else {
			int c=0;
			while (c<length && (r=in.read())>-1){
				sbResponse.append((char)r);
				bytes++;
				c++;
			}
		}
	}
	
	private int readChunkSize(int b0) throws IOException {
		int chunk=0;
		int b=b0;
		while (b!=13) {
			chunk=addChunkChar(chunk, (char) b);
			b=in.read();
		}
		in.skip(1);	
		return chunk;
	}
	
//	private void changeState() {
//		synchronized (cancelLock) {
//			active=!active;
//		}
//	}
	
	private void setActive(){
		synchronized (cancelLock) {
			active=true;
		}
	}
	
	private void setInactive(){
		synchronized (cancelLock) {
			active=false;
//			cancelLock.notify();
		}
	}
	
	public void removeRequestProperty(String name){
		requestProperties.remove(name);
	}
	
	public void setGzipEnabled(boolean gzipEnabled) {
		this.gzipEnabled = gzipEnabled;
	}
	
	public boolean isGzipEnabled() {
		return gzipEnabled;
	}
}
